#include "../../internal-include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/dma_io.hpp>
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.h>

#include <stdio.h>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            __declare_io_type(Mode, uint8_t, 
                static constexpr auto DoubleSpeed    = Bit(7);
                static constexpr auto SingleSpeed    = !DoubleSpeed;
                static constexpr auto XADPCM         = Bit(6);
                static constexpr auto WholeSector    = Bit(5);
                static constexpr auto DataSector     = !WholeSector;
                static constexpr auto UseXAFilter    = Bit(3);
                static constexpr auto AudioPlayIRQ   = Bit(2);
                static constexpr auto AutoPauseTrack = Bit(1);
                static constexpr auto CDDA           = Bit(0);
            );

            static constexpr auto DataSectorMode = Mode_t::from(Mode_t::DoubleSpeed, Mode_t::DataSector);
            
            static SectorBufferAllocator sector_allocator;
            static uint16_t              sectors_left;

            static void pause_cd() {
                CD_IO::PortIndex0::change_to();
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::Pause);
            }

            static void read_sector_dma(CD_IO::DataSector& sector) {
                static const auto WaitSectorReady = []() {
                    while(!CD_IO::IndexStatus.is_set(CD_IO::IndexStatus_t::HasDataFifoData));
                };

                static const auto ReadSector = [](uint32_t* dst) {
                    DMA_IO::CDROM.set_adr(reinterpret_cast<uintptr_t>(dst));
                    DMA_IO::CDROM.block_ctrl   = DMA_IO::BCR_t::SyncMode0::for_cd();
                    DMA_IO::CDROM.channel_ctrl = DMA_IO::CHCHR_t::StartCDROM();
                    
                    DMA_IO::CDROM.wait();

                    CD_IO::PortIndex0::Request.reset();
                };

                WaitSectorReady();
                ReadSector(sector.data);
            }

            static void read_sector_to(CD_IO::DataSector& sector) {
                CD_IO::PortIndex0::change_to();
                CD_IO::PortIndex0::Request.want_data();

                // We only support DMA rn
                read_sector_dma(sector);

                // Do we ever want to support reading via IO Port?
                // Doesn't seem to important when we can use DMA
            }

            static bool try_read_sector() {
                // Obtain sector content here
                auto* sector = sector_allocator.allocate_sector();
                if(sector) {
                    //Now obtain sector
                    read_sector_to(*sector);

                    sectors_left--;
                    if(sectors_left == 0) {
                        current_state = State::Done;
                        pause_cd();
                    }
                    return true;
                }
                return false;
            }

            static InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::CDROM)) {
                    const uint8_t old_status = CD_IO::IndexStatus;

                    CD_IO::PortIndex1::change_to();
                    const auto cur_irq = CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlag);
                    last_interrupt = cur_irq;
                    CD_IO::Interrupt::ack(CD_IO::PortIndex1::InterruptFlag);

                    if(cur_irq == CD_IO::Interrupt::DataReady) {
                        //Obtain sector content here
                        if(!try_read_sector()) {
                            current_state = State::BufferFull;
                            pause_cd();
                        }
                    }

                    else if(cur_irq == CD_IO::Interrupt::DiskError) {
                        current_state = State::Error;
                    }

                    // No masking required because we can only write bit 0 - 2
                    CD_IO::IndexStatus = old_status;
                    return InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return InterruptVerifierResult::SkipHandler;
                }
            }

            static void interrupt_handler(uint32_t) {
                Interrupt::ack_irq(Interrupt::CDROM);
                __syscall_ReturnFromException();
            }

            CD_IO::Interrupt::Type last_interrupt = CD_IO::Interrupt::Type::None;
            State                  current_state  = State::Free;
            InterrupCallback callback = {
                .next              = nullptr,
                .handler_function  = reinterpret_cast<InterruptHandler>(interrupt_handler),
                .verifier_function = interrupt_verifier
            };

            void read_file(FileInfo file_info, const SectorBufferAllocator& buffer_allocator) {
                sector_allocator = buffer_allocator;
                sectors_left     = file_info.sectors;

                CD_IO::PortIndex0::change_to();
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetMode, DataSectorMode);

                const auto loc = CDTimeStamp::from(file_info);
                Command::send_wait<CD_IO::PortIndex0>(CD_IO::Command::SetLoc, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
                Command::send<CD_IO::PortIndex0>(CD_IO::Command::ReadN);

                current_state = State::Reading;

                printf("Now reading: %i(%i:%i:%i)\n", file_info.lba, loc.get_min_cd(), loc.get_sec_cd(), loc.get_sector_cd());
            }

            bool continue_reading() {
                if(current_state != State::BufferFull) {
                    return true;
                }

                CD_IO::PortIndex1::change_to();
                return try_read_sector();
            }
        }
    }
}